/**
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.pxf.plugins.tarantool.insert;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.codehaus.jackson.map.ObjectMapper;
import org.greenplum.pxf.api.OneRow;
import org.greenplum.pxf.api.model.Accessor;
import org.greenplum.pxf.api.model.BasePlugin;
import org.greenplum.pxf.api.model.RequestContext;
import org.greenplum.pxf.api.utilities.ColumnDescriptor;
import ru.datamart.pxf.plugins.tarantool.client.HttpClientProvider;
import ru.datamart.pxf.plugins.tarantool.client.HttpClientProviderImpl;
import ru.datamart.pxf.plugins.tarantool.dto.AdgQueryRequest;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

public class TarantoolAccessor extends BasePlugin implements Accessor {
    private static final int DEFAULT_TIMEOUT_CONNECT = 5000;
    private static final int DEFAULT_TIMEOUT_READ = 5000;
    private static final int DEFAULT_TIMEOUT_REQUEST = 5000;
    private static final int DEFAULT_BUFFER_SIZE = 5000;
    private static final String ENDPOINT = "tarantool.rest.endpoint";
    private static final String TIMEOUT_CONNECT = "tarantool.rest.timeout.connect";
    private static final String TIMEOUT_READ = "tarantool.rest.timeout.read";
    private static final String TIMEOUT_REQUEST = "tarantool.rest.timeout.request";
    private static final String BUFFER_SIZE = "tarantool.rest.buffer.size";
    private static final String INSERT_INTO_TEMPLATE = "INSERT INTO \"%s\" (%s) VALUES %s";
    private final ObjectMapper objectMapper = new ObjectMapper();
    private HttpClientProvider httpClientProvider = new HttpClientProviderImpl();
    private String spaceName;
    private String endpoint;
    private int bufferSize = DEFAULT_BUFFER_SIZE;
    private CloseableHttpClient httpClient;
    private int connectTimeout;
    private int readTimeout;
    private int requestTimeout;
    private List<List<?>> currentBatch;
    private String columns;

    @Override
    public void initialize(RequestContext requestContext) {
        super.initialize(requestContext);

        spaceName = requestContext.getDataSource();
        if (StringUtils.isBlank(spaceName)) {
            throw new IllegalArgumentException("Tarantool space must be set");
        }

        endpoint = configuration.get(ENDPOINT);
        if (StringUtils.isBlank(endpoint)) {
            throw new IllegalArgumentException("TARANTOOL_SERVER property must be set");
        }

        String bufferSize = configuration.get(BUFFER_SIZE);
        if (StringUtils.isNotBlank(bufferSize)) {
            this.bufferSize = Integer.parseInt(bufferSize);
        }

        StringBuilder columns = new StringBuilder();
        for (int i = 0; i < requestContext.getColumns(); i++) {
            ColumnDescriptor column = requestContext.getColumn(i);
            if (i > 0) {
                columns.append(",");
            }
            columns.append('"').append(column.columnName()).append('"');
        }
        this.columns = columns.toString();

        this.connectTimeout = configuration.getInt(TIMEOUT_CONNECT, DEFAULT_TIMEOUT_CONNECT);
        this.readTimeout = configuration.getInt(TIMEOUT_READ, DEFAULT_TIMEOUT_READ);
        this.requestTimeout = configuration.getInt(TIMEOUT_REQUEST, DEFAULT_TIMEOUT_REQUEST);
        this.currentBatch = new ArrayList<>(this.bufferSize);
    }

    @Override
    public boolean openForWrite() throws Exception {
        try {
            LOG.info("Opening [{}] for write in [{}]. Segment: {}, total: {}",
                    context.getProfile(), spaceName, context.getSegmentId(), context.getTotalSegments());
            httpClient = httpClientProvider.provide(connectTimeout, readTimeout, requestTimeout);
            return true;
        } catch (Throwable e) {
            LOG.error("Failed opening [{}] for write in [{}]. Segment: {}, total: {}",
                    context.getProfile(), spaceName, context.getSegmentId(), context.getTotalSegments(), e);
            closeHttpClientIfOpened();
            throw e;
        }
    }

    @Override
    public boolean writeNextObject(OneRow oneRow) throws Exception {
        List<?> columns = (List<?>) oneRow.getData();
        currentBatch.add(columns);
        if (currentBatch.size() >= bufferSize) {
            sendBatch();
        }

        return true;
    }

    @Override
    public void closeForWrite() throws Exception {
        try {
            LOG.info("Closing [{}] for write in [{}]. Batch: {}, segment: {}, total: {}",
                    context.getProfile(), spaceName, currentBatch.size(), context.getSegmentId(), context.getTotalSegments());
            sendBatch();
        } finally {
            currentBatch.clear();
            closeHttpClientIfOpened();
            LOG.info("Closed [{}] for write in [{}]. Segment: {}, total: {}",
                    context.getProfile(), spaceName, context.getSegmentId(), context.getTotalSegments());
        }
    }

    @Override
    public boolean openForRead() throws Exception {
        throw new UnsupportedOperationException();
    }

    @Override
    public OneRow readNextObject() throws Exception {
        throw new UnsupportedOperationException();
    }

    @Override
    public void closeForRead() throws Exception {
        throw new UnsupportedOperationException();
    }

    public void setHttpClientProvider(HttpClientProvider httpClientProvider) {
        this.httpClientProvider = httpClientProvider;
    }

    private void sendBatch() throws IOException {
        if (currentBatch.isEmpty()) {
            return;
        }

        LOG.info("Sending batch [{}] for write in [{}]. Batch: {} segment: {}, total: {}",
                context.getProfile(), spaceName, currentBatch.size(), context.getSegmentId(), context.getTotalSegments());
        int rowSize = currentBatch.get(0).size();
        String valuesOfRow = StringUtils.repeat("?", ",", rowSize);
        String row = String.format("(%s)", valuesOfRow);
        String values = StringUtils.repeat(row, ",", currentBatch.size());
        String sqlBatch = String.format(INSERT_INTO_TEMPLATE, spaceName, columns, values);
        List<Object> params = currentBatch.stream()
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        currentBatch.clear();

        HttpPost httpPost = new HttpPost(endpoint);
        String queryRequestJson = objectMapper.writeValueAsString(new AdgQueryRequest(sqlBatch, params));
        httpPost.setEntity(new StringEntity(queryRequestJson, ContentType.APPLICATION_JSON));
        try (CloseableHttpResponse response = httpClient.execute(httpPost)) {
            int statusCode = response.getStatusLine().getStatusCode();
            if (HttpStatus.SC_OK != statusCode) {
                String errorMessage = readErrorMessage(response);
                LOG.error("Adg respond with error [{}] for write in [{}]. Segment: {}, total: {}, status: {}, error: [{}]",
                        context.getProfile(), spaceName, context.getSegmentId(), context.getTotalSegments(), statusCode, errorMessage);
                throw new IllegalStateException(String.format("Adg respond with error [%s]: [%s]", statusCode, errorMessage));
            }
            LOG.debug("Batch sent [{}] for write in [{}]. Segment: {}, total: {}",
                    context.getProfile(), spaceName, context.getSegmentId(), context.getTotalSegments());
        }
    }

    private String readErrorMessage(CloseableHttpResponse response) {
        StringBuilder stringBuilder = new StringBuilder();
        try (BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(response.getEntity().getContent()))) {
            String line;
            while ((line = bufferedReader.readLine()) != null) {
                stringBuilder.append(line).append("\n");
            }
            return stringBuilder.toString();
        } catch (Exception e) {
            LOG.error("Failed read response [{}] for write in [{}]. Segment: {}, total: {}",
                    context.getProfile(), spaceName, context.getSegmentId(), context.getTotalSegments(), e);
            return "Failed to read response";
        }
    }

    private void closeHttpClientIfOpened() {
        try {
            if (httpClient != null) {
                httpClient.close();
                httpClient = null;
            }
        } catch (IOException e) {
            LOG.warn("Could not close http client", e);
        }
    }
}
