/**
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.pxf.plugins.tarantool.insert;

import org.apache.http.HttpVersion;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.message.BasicStatusLine;
import org.greenplum.pxf.api.OneRow;
import org.greenplum.pxf.api.io.DataType;
import org.greenplum.pxf.api.model.RequestContext;
import org.greenplum.pxf.api.utilities.ColumnDescriptor;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import ru.datamart.pxf.plugins.tarantool.client.HttpClientProvider;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;

import static java.util.Arrays.asList;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
class TarantoolAccessorTest {
    private static final String USERNAME = "someuser";
    private static final String ENDPOINT = "localhost:1111";
    private static final String PASSWORD = "123";
    private static final String TIMEOUT_CONNECT = "4000";
    private static final String TIMEOUT_READ = "5000";
    private static final String TIMEOUT_REQUEST = "6000";
    private static final String SPACE = "space";
    public static final String BUFFER_SIZE = "2";

    @Mock
    private HttpClientProvider httpClientProvider;

    @Mock
    private CloseableHttpClient closeableHttpClient;

    @Mock
    private CloseableHttpResponse response;

    private RequestContext context;

    private TarantoolAccessor tarantoolDeleteAccessor;

    @BeforeEach
    void setUp() throws IOException {
        context = new RequestContext();
        context.setSegmentId(0);
        context.setTotalSegments(1);
        context.setUser("user");
        context.setConfig("default");
        context.setDataSource(SPACE);
        context.setTupleDescription(Arrays.asList(
                new ColumnDescriptor("id", DataType.BIGINT.getOID(), 0, null, null),
                new ColumnDescriptor("name", DataType.VARCHAR.getOID(), 0, null, null),
                new ColumnDescriptor("bucket_id", DataType.BIGINT.getOID(), 0, null, null)));
        context.setAdditionalConfigProps(new HashMap<String, String>() {{
            put("tarantool.rest.endpoint", ENDPOINT);
            put("tarantool.rest.user", USERNAME);
            put("tarantool.rest.password", PASSWORD);
            put("tarantool.rest.timeout.connect", TIMEOUT_CONNECT);
            put("tarantool.rest.timeout.read", TIMEOUT_READ);
            put("tarantool.rest.timeout.request", TIMEOUT_REQUEST);
            put("tarantool.rest.buffer.size", BUFFER_SIZE);
        }});

        lenient().when(httpClientProvider.provide(anyInt(), anyInt(), anyInt())).thenReturn(closeableHttpClient);
        lenient().when(closeableHttpClient.execute(any())).thenReturn(response);

        tarantoolDeleteAccessor = new TarantoolAccessor();
        tarantoolDeleteAccessor.setHttpClientProvider(httpClientProvider);
    }

    @Test
    void shouldFailWhenNoSpaceDefined() {
        // arrange
        context.setDataSource("");

        // act assert
        assertThrows(IllegalArgumentException.class, () -> tarantoolDeleteAccessor.initialize(context));
    }

    @Test
    void shouldFailWhenNoHostDefined() {
        // arrange
        context.getAdditionalConfigProps().remove("tarantool.rest.endpoint");

        // act assert
        assertThrows(IllegalArgumentException.class, () -> tarantoolDeleteAccessor.initialize(context));
    }

    @Test
    void shouldCorrectlyProcess() throws Exception {
        // act
        when(response.getStatusLine()).thenReturn(new BasicStatusLine(new HttpVersion(1, 1), 200, ""));

        tarantoolDeleteAccessor.initialize(context);
        assertTrue(tarantoolDeleteAccessor.openForWrite());
        assertTrue(tarantoolDeleteAccessor.writeNextObject(new OneRow(Arrays.asList(1L, "test", 1))));
        assertTrue(tarantoolDeleteAccessor.writeNextObject(new OneRow(Arrays.asList(2L, "test2", 2))));
        assertTrue(tarantoolDeleteAccessor.writeNextObject(new OneRow(Arrays.asList(3L, "test3", 3))));
        assertTrue(tarantoolDeleteAccessor.writeNextObject(new OneRow(Arrays.asList(4L, "test4", 4))));
        assertTrue(tarantoolDeleteAccessor.writeNextObject(new OneRow(Arrays.asList(5L, "test5", 5))));
        tarantoolDeleteAccessor.closeForWrite();

        // assert
        verify(closeableHttpClient, times(3)).execute(any());

    }

    @Test
    void shouldFailWhenExceptionDuringExecutingRequestOnBatch() throws Exception {
        // arrange
        when(closeableHttpClient.execute(any())).thenThrow(new RuntimeException("Failed"));

        // act
        tarantoolDeleteAccessor.initialize(context);
        tarantoolDeleteAccessor.openForWrite();
        tarantoolDeleteAccessor.writeNextObject(new OneRow(asList(1L, "test", 1)));
        assertThrows(RuntimeException.class, () -> tarantoolDeleteAccessor.writeNextObject(new OneRow(asList(1L, "test", 1))));
        tarantoolDeleteAccessor.closeForWrite();

        // asssert
        verify(closeableHttpClient, times(1)).execute(any());
    }
}
